class EndPoints:
    TICKETS = "/api/v2/tickets"


class Status:
    OPEN = 2
    PENDING = 3
    RESOLVED = 4
    CLOSED = 5


class Source:
    EMAIL = 1
    PORTAL = 2


class Priority:
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    URGENT = 4



from pathlib import Path

from cg.constants.constants import MicrosaltQC
from cg.meta.workflow.microsalt.constants import QUALITY_REPORT_FILE_NAME
from cg.meta.workflow.microsalt.metrics_parser.models import SampleMetrics
from cg.meta.workflow.microsalt.quality_controller.models import (
    CaseQualityResult,
    SampleQualityResult,
)
from cg.meta.workflow.microsalt.quality_controller.result_logger import ResultLogger
from cg.models.orders.sample_base import ControlEnum
from cg.store.models import Sample


def is_valid_total_reads(reads: int, target_reads: int, threshold_percentage: int) -> bool:
    return reads > target_reads * threshold_percentage / 100


def is_valid_total_reads_for_negative_control(reads: int, target_reads: int) -> bool:
    return reads < target_reads * MicrosaltQC.NEGATIVE_CONTROL_READS_THRESHOLD


def is_valid_mapping_rate(mapping_rate: float) -> bool:
    return mapping_rate > MicrosaltQC.MAPPED_RATE_THRESHOLD


def is_valid_duplication_rate(duplication_rate: float) -> bool:
    return duplication_rate < MicrosaltQC.DUPLICATION_RATE_THRESHOLD


def is_valid_median_insert_size(insert_size: int) -> bool:
    return insert_size > MicrosaltQC.INSERT_SIZE_THRESHOLD


def is_valid_average_coverage(average_coverage: float) -> bool:
    return average_coverage > MicrosaltQC.AVERAGE_COVERAGE_THRESHOLD


def is_valid_10x_coverage(coverage_10x: float) -> bool:
    return coverage_10x > MicrosaltQC.COVERAGE_10X_THRESHOLD


def has_valid_mapping_rate(metrics: SampleMetrics) -> bool:
    mapped_rate: float | None = metrics.microsalt_samtools_stats.mapped_rate
    return is_valid_mapping_rate(mapped_rate) if mapped_rate else False


def has_valid_duplication_rate(metrics: SampleMetrics) -> bool:
    duplication_rate: float | None = metrics.picard_markduplicate.duplication_rate
    return is_valid_duplication_rate(duplication_rate) if duplication_rate else False


def has_valid_median_insert_size(metrics: SampleMetrics) -> bool:
    insert_size: int | None = metrics.picard_markduplicate.insert_size
    return is_valid_median_insert_size(insert_size) if insert_size else False


def has_valid_average_coverage(metrics: SampleMetrics) -> bool:
    coverage: float | None = metrics.microsalt_samtools_stats.average_coverage
    return is_valid_average_coverage(coverage) if coverage else False


def has_valid_10x_coverage(metrics: SampleMetrics) -> bool:
    coverage_10x: float | None = metrics.microsalt_samtools_stats.coverage_10x
    return is_valid_10x_coverage(coverage_10x) if coverage_10x else False


def get_negative_control_result(results: list[SampleQualityResult]) -> SampleQualityResult | None:
    for result in results:
        if result.is_control:
            return result


def negative_control_pass_qc(results: list[SampleQualityResult]) -> bool:
    if negative_control_result := get_negative_control_result(results):
        return negative_control_result.passes_qc
    return True


def is_sample_negative_control(sample: Sample) -> bool:
    return sample.control == ControlEnum.negative


def get_application_tag(sample: Sample) -> str:
    return sample.application_version.application.tag


def get_sample_target_reads(sample: Sample) -> int:
    return sample.application_version.application.target_reads


def get_percent_reads_guaranteed(sample: Sample) -> int:
    return sample.application_version.application.percent_reads_guaranteed


def get_report_path(metrics_file_path: Path) -> Path:
    return metrics_file_path.parent.joinpath(QUALITY_REPORT_FILE_NAME)


def quality_control_case(sample_results: list[SampleQualityResult]) -> CaseQualityResult:
    control_pass_qc: bool = negative_control_pass_qc(sample_results)
    case_passes_qc: bool = all(sample.passes_qc for sample in sample_results)
    result = CaseQualityResult(
        passes_qc=case_passes_qc,
        control_passes_qc=control_pass_qc,
    )
    ResultLogger.log_case_result(result)
    return result



import logging
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path

from cg.apps.environ import environ_email
from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants.constants import CaseActions, CustomerId, Workflow
from cg.constants.priority import TrailblazerPriority
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.constants.tb import AnalysisType
from cg.exc import AnalysisRunningError
from cg.meta.workflow.utils.utils import MAP_TO_TRAILBLAZER_PRIORITY
from cg.services.analysis_starter.configurator.abstract_model import CaseConfig
from cg.store.models import Analysis, Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class Tracker(ABC):
    """Ensures tracking of started analyses. This mainly means exporting analyses to Trailblazer."""

    def __init__(
        self,
        store: Store,
        trailblazer_api: TrailblazerAPI,
        workflow_root: str,
    ):
        self.store = store
        self.trailblazer_api = trailblazer_api
        self.workflow_root = workflow_root

    def track(
        self,
        case_config: CaseConfig,
    ) -> None:
        tb_analysis: TrailblazerAnalysis = self._track_in_trailblazer(
            case_id=case_config.case_id, tower_workflow_id=case_config.get_workflow_id()
        )
        LOG.info(
            f"Analysis entry for case {case_config.case_id} created in Trailblazer with id {tb_analysis.id}"
        )
        self._create_analysis_statusdb(case_config=case_config, trailblazer_id=tb_analysis.id)
        LOG.info(f"Analysis entry for case {case_config.case_id} created in StatusDB")

    def ensure_analysis_not_ongoing(self, case_id: str) -> None:
        """Raises: AnalysisRunningError if the case has an analysis running in Trailblazer."""
        if self.trailblazer_api.is_latest_analysis_ongoing(case_id):
            raise AnalysisRunningError(f"Analysis still ongoing in Trailblazer for case {case_id}")
        LOG.info(f"No ongoing analysis in Trailblazer for case {case_id}")

    def set_case_as_running(self, case_id: str) -> None:
        self.store.update_case_action(case_internal_id=case_id, action=CaseActions.RUNNING)

    def set_case_as_not_running(self, case_id: str) -> None:
        self.store.update_case_action(case_internal_id=case_id, action=None)

    def _track_in_trailblazer(
        self, case_id: str, tower_workflow_id: str | None
    ) -> TrailblazerAnalysis:
        analysis_type: str = self._get_analysis_type(case_id)
        config_path: Path = self._get_job_ids_path(case_id)
        email: str = environ_email()
        order_id: int = self.store.get_case_by_internal_id_strict(case_id).latest_order.id
        out_dir: str = config_path.parent.as_posix()
        priority: TrailblazerPriority = self._get_trailblazer_priority(case_id)
        ticket: str = self.store.get_latest_ticket_from_case(case_id)
        is_case_for_development: bool = self._is_case_for_development(case_id)
        return self.trailblazer_api.add_pending_analysis(
            analysis_type=analysis_type,
            case_id=case_id,
            config_path=config_path.as_posix(),
            email=email,
            order_id=order_id,
            out_dir=out_dir,
            priority=priority,
            ticket=ticket,
            workflow=self.store.get_case_workflow(case_id),
            workflow_manager=self._workflow_manager(),
            tower_workflow_id=tower_workflow_id,
            is_hidden=is_case_for_development,
        )

    def _create_analysis_statusdb(
        self, case_config: CaseConfig, trailblazer_id: int | None
    ) -> None:
        """Storing an analysis bundle in StatusDB for a provided case."""
        case_id: str = case_config.case_id
        LOG.info(f"Storing analysis in StatusDB for {case_id}")
        case: Case = self.store.get_case_by_internal_id_strict(case_id)
        is_primary: bool = len(case.analyses) == 0
        analysis_start: datetime = datetime.now()
        workflow_version: str = self._get_workflow_version(case_config)
        new_analysis: Analysis = self.store.add_analysis(
            case=case,
            workflow=Workflow(case.data_analysis),
            version=workflow_version,
            completed_at=None,
            primary=is_primary,
            session_id=case_config.get_session_id(),
            started_at=analysis_start,
            trailblazer_id=trailblazer_id,
        )
        self.store.add_item_to_store(new_analysis)
        self.store.commit_to_store()
        LOG.info(f"Analysis successfully stored in StatusDB: {case_id} : {analysis_start}")

    def _get_analysis_type(self, case_id: str) -> str:
        """
        Return the analysis type for sample.
        Only analysis types supported by Trailblazer are valid outputs.
        """
        sample: Sample = self.store.get_case_by_internal_id_strict(case_id).samples[0]
        prep_category: str = sample.prep_category
        if prep_category and prep_category.lower() in {
            SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
            SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING,
            SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
            SeqLibraryPrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING,
        }:
            return prep_category.lower()
        return AnalysisType.OTHER

    def _get_trailblazer_priority(self, case_id: str) -> TrailblazerPriority:
        """Get the priority for the case in Trailblazer."""
        case: Case = self.store.get_case_by_internal_id_strict(internal_id=case_id)
        return MAP_TO_TRAILBLAZER_PRIORITY[case.priority]

    def _is_case_for_development(self, case_id: str) -> bool:
        case: Case = self.store.get_case_by_internal_id_strict(case_id)
        return case.customer.internal_id == CustomerId.CG_INTERNAL_CUSTOMER

    @abstractmethod
    def _workflow_manager(self):
        pass

    @abstractmethod
    def _get_job_ids_path(self, case_id: str):
        pass

    @abstractmethod
    def _get_workflow_version(self, case_config: CaseConfig) -> str:
        pass



import logging
from pathlib import Path
from typing import cast

from cg.constants.priority import SlurmQos
from cg.exc import CaseNotConfiguredError
from cg.meta.workflow.fastq import BalsamicFastqHandler
from cg.models.cg_config import BalsamicConfig
from cg.services.analysis_starter.configurator.configurator import Configurator
from cg.services.analysis_starter.configurator.file_creators.balsamic_config import (
    BalsamicConfigFileCreator,
)
from cg.services.analysis_starter.configurator.models.balsamic import BalsamicCaseConfig
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class BalsamicConfigurator(Configurator):
    def __init__(
        self,
        config: BalsamicConfig,
        config_file_creator: BalsamicConfigFileCreator,
        fastq_handler: BalsamicFastqHandler,
        store: Store,
    ):
        self.store: Store = store

        self.balsamic_binary: Path = config.binary_path
        self.conda_binary: Path = config.conda_binary
        self.environment: str = config.conda_env
        self.head_job_partition: str = config.head_job_partition
        self.root_dir: Path = config.root
        self.slurm_account: str = config.slurm.account

        self.fastq_handler: BalsamicFastqHandler = fastq_handler
        self.config_file_creator: BalsamicConfigFileCreator = config_file_creator

    def configure(self, case_id: str, **flags) -> BalsamicCaseConfig:
        LOG.info(f"Configuring case {case_id}")
        self.fastq_handler.link_fastq_files(case_id)
        fastq_path: Path = self.fastq_handler.get_fastq_dir(case_id)
        self.config_file_creator.create(case_id=case_id, fastq_path=fastq_path, **flags)
        return self.get_config(case_id=case_id, **flags)

    def get_config(self, case_id: str, **flags) -> BalsamicCaseConfig:
        balsamic_config: BalsamicCaseConfig = BalsamicCaseConfig(
            account=self.slurm_account,
            binary=self.balsamic_binary,
            case_id=case_id,
            conda_binary=self.conda_binary,
            environment=self.environment,
            head_job_partition=self.head_job_partition,
            qos=cast(SlurmQos, self.store.get_case_by_internal_id_strict(case_id).slurm_priority),
            sample_config=self._get_sample_config_path(case_id),
        )
        balsamic_config: BalsamicCaseConfig = self._set_flags(config=balsamic_config, **flags)
        self._ensure_required_config_files_exist(balsamic_config)
        return balsamic_config

    def _get_sample_config_path(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id, f"{case_id}.json")

    def _ensure_required_config_files_exist(self, config: BalsamicCaseConfig) -> None:
        if not config.sample_config.exists():
            raise CaseNotConfiguredError(
                f"Please ensure that the config file {config.sample_config} exists."
            )



import logging
import re

from cg.constants.demultiplexing import (
    FORWARD_INDEX_CYCLE_PATTERN,
    REVERSE_INDEX_CYCLE_PATTERN,
    SampleSheetBCLConvertSections,
)
from cg.exc import OverrideCyclesError

LOG = logging.getLogger(__name__)


class OverrideCyclesValidator:
    """Class for validating the override cycles value of a sample in a sample sheet."""

    def __init__(
        self,
        run_read1_cycles: int | None = None,
        run_read2_cycles: int | None = None,
        run_index1_cycles: int | None = None,
        run_index2_cycles: int | None = None,
        is_reverse_complement: bool | None = None,
    ):
        self.sample: dict[str, str] | None = None
        self.sample_cycles: list[str] | None = None
        self.sample_id: str | None = None
        self.run_read1_cycles: int | None = run_read1_cycles
        self.run_read2_cycles: int | None = run_read2_cycles
        self.run_index1_cycles: int | None = run_index1_cycles
        self.run_index2_cycles: int | None = run_index2_cycles
        self.is_reverse_complement: bool | None = is_reverse_complement

    def set_attributes_from_sample(self, sample: dict[str, str]) -> None:
        """Set the sample, override cycles and sample id."""
        self.sample = sample
        self.sample_cycles = sample[SampleSheetBCLConvertSections.Data.OVERRIDE_CYCLES].split(";")
        self.sample_id = sample[SampleSheetBCLConvertSections.Data.SAMPLE_INTERNAL_ID]

    @staticmethod
    def _is_index_cycle_value_following_pattern(
        pattern: str, index_cycle: str, run_cycles: int, index_sequence: str
    ) -> bool:
        """
        Returns whether an index cycle string is following a valid cycle regex pattern and has
        consistent values. Valid patterns are 'I(\d+)N(\d+)' and 'N(\d+)I(\d+)'. Having consistent
        values means that the sum of the number of index characters (I) and the number of ignored
        characters (N) specified in the index cycle string is equal to the number of run cycles
        and the length of the index sequence is equal to the number of index characters (I).
        """
        match = re.fullmatch(pattern, index_cycle)
        if match:
            if pattern == FORWARD_INDEX_CYCLE_PATTERN:
                index_chars, ignored_chars = map(int, match.groups())
            elif pattern == REVERSE_INDEX_CYCLE_PATTERN:
                ignored_chars, index_chars = map(int, match.groups())
            else:
                LOG.warning(f"Pattern {pattern} is not a valid index cycle pattern")
                return False
            if index_chars + ignored_chars == run_cycles and len(index_sequence) == index_chars:
                return True
        return False

    def _validate_reads_cycles(self) -> None:
        """
        Determine if the sample read cycles are valid, i.e. if the sample read cycle values are
        equal to the respective run read cycles.
        Raises:
            OverrideCyclesError if the reads cycles are not valid.
        """
        read1_cycle: str = self.sample_cycles[0]
        read2_cycle: str = self.sample_cycles[-1]
        if (
            read1_cycle == f"Y{self.run_read1_cycles}"
            and read2_cycle == f"Y{self.run_read2_cycles}"
        ):
            return
        message: str = f"Incorrect read cycles {self.sample_cycles} for sample {self.sample_id}"
        LOG.error(message)
        raise OverrideCyclesError(message)

    def _validate_index1_cycles(self) -> None:
        """
        Determine if the sample index 1 cycle is valid, i.e., if the number of index cycles is
        coherent with the run cycles and the index sequence length.
        Raises:
            OverrideCyclesError if the index 1 cycle is not valid.
        """
        index1_cycle: str = self.sample_cycles[1]
        if (
            self.run_index1_cycles == len(self.sample[SampleSheetBCLConvertSections.Data.INDEX_1])
            and index1_cycle == f"I{self.run_index1_cycles}"
        ):
            return
        if self._is_index_cycle_value_following_pattern(
            pattern=FORWARD_INDEX_CYCLE_PATTERN,
            index_cycle=index1_cycle,
            run_cycles=self.run_index1_cycles,
            index_sequence=self.sample[SampleSheetBCLConvertSections.Data.INDEX_1],
        ):
            return
        message: str = f"Incorrect index1 cycle {index1_cycle} for sample {self.sample_id}"
        LOG.error(message)
        raise OverrideCyclesError(message)

    def _validate_index2_cycles(self) -> None:
        """
        Determine if the index 2 cycle is valid, i.e., if the number of ignored and index characters
        correspond to the length of the sample index2 sequence and the number of run index2 cycles,
        or if the index cycles should be None.
        Raises:
            OverrideCyclesError if the index 2 cycle is not valid.
        """
        if not self.run_index2_cycles and len(self.sample_cycles) == 3:
            return
        index2_cycle: str = self.sample_cycles[2]
        if (
            not self.sample[SampleSheetBCLConvertSections.Data.INDEX_2]
            and index2_cycle == f"N{self.run_index2_cycles}"
        ):
            return
        if (
            self.run_index2_cycles == len(self.sample[SampleSheetBCLConvertSections.Data.INDEX_2])
            and index2_cycle == f"I{self.run_index2_cycles}"
        ):
            return
        if self.is_reverse_complement and self._is_index_cycle_value_following_pattern(
            pattern=REVERSE_INDEX_CYCLE_PATTERN,
            index_cycle=index2_cycle,
            run_cycles=self.run_index2_cycles,
            index_sequence=self.sample[SampleSheetBCLConvertSections.Data.INDEX_2],
        ):
            return
        if not self.is_reverse_complement and self._is_index_cycle_value_following_pattern(
            pattern=FORWARD_INDEX_CYCLE_PATTERN,
            index_cycle=index2_cycle,
            run_cycles=self.run_index2_cycles,
            index_sequence=self.sample[SampleSheetBCLConvertSections.Data.INDEX_2],
        ):
            return
        message: str = (
            f"Index2 cycle {index2_cycle} of sample {self.sample_id} "
            f"does not match with run cycle {self.run_index2_cycles} "
            f"and reverse complement set to {self.is_reverse_complement}"
        )
        LOG.error(message)
        raise OverrideCyclesError(message)

    def validate_sample(
        self,
        sample: dict[str, str],
    ) -> None:
        """Determine if the override cycles are valid for a given sample."""
        self.set_attributes_from_sample(sample)
        self._validate_reads_cycles()
        self._validate_index1_cycles()
        self._validate_index2_cycles()



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.store.models import Invoice


def filter_invoices_by_invoice_id(invoices: Query, entry_id: int, **kwargs) -> Query:
    """Return invoices by invoice id."""
    return invoices.filter(Invoice.id == entry_id)


def filter_invoices_invoiced(invoices: Query, **kwargs) -> Query:
    """Return invoices by invoiced at."""
    return invoices.filter(Invoice.invoiced_at.isnot(None))


def filter_invoices_not_invoiced(invoices: Query, **kwargs) -> Query:
    """Return invoices by not invoiced at."""
    return invoices.filter(Invoice.invoiced_at.is_(None))


def apply_invoice_filter(
    filter_functions: list[str],
    invoices: Query,
    entry_id: int | None = None,
) -> Query:
    """Apply filtering functions to the invoice queries and return filtered results."""

    for filter_function in filter_functions:
        invoices: Query = filter_function(invoices=invoices, entry_id=entry_id)
    return invoices


class InvoiceFilter(Enum):
    """Define Invoice filter functions."""

    BY_INVOICE_ID: Callable = filter_invoices_by_invoice_id
    BY_INVOICED: Callable = filter_invoices_invoiced
    BY_NOT_INVOICED: Callable = filter_invoices_not_invoiced



from enum import StrEnum

OSTYPE: str = "Unix/MacOS"
ROOT_TO_TRIM: str = "/home"

DESTINATION_ATTRIBUTE: str = "destination"
SOURCE_ATTRIBUTE: str = "source"

DELETE_FILE_SUCCESSFUL_MESSAGE: str = "Object has been deleted"


class DataflowEndpoints(StrEnum):
    """Enum containing all DDN dataflow endpoints used."""

    ARCHIVE_FILES = "files/archive"
    DELETE_FILE = "files/delete"
    GET_AUTH_TOKEN = "auth/token"
    REFRESH_AUTH_TOKEN = "auth/token/refresh"
    RETRIEVE_FILES = "files/retrieve"
    GET_JOB_STATUS = "activity/jobs/"


class JobStatus(StrEnum):
    """Enum for the different job statuses which can be returned via Miria."""

    CANCELED = "Canceled"
    COMPLETED = "Completed"
    DENIED = "Denied"
    CREATION_IN_PROGRESS = "Creation in progress"
    IN_QUEUE = "In Queue"
    INVALID_LICENSE = "Invalid license"
    ON_VALIDATION = "On validation"
    REFUSED = "Refused"
    RUNNING = "Running"
    SUSPENDED = "Suspended"
    TERMINATED_ON_ERROR = "Terminated on error"
    TERMINATED_ON_WARNING = "Terminated on warning"


FAILED_JOB_STATUSES: list[str] = [
    JobStatus.CANCELED,
    JobStatus.DENIED,
    JobStatus.INVALID_LICENSE,
    JobStatus.REFUSED,
    JobStatus.TERMINATED_ON_ERROR,
    JobStatus.TERMINATED_ON_WARNING,
]
ONGOING_JOB_STATUSES: list[str] = [
    JobStatus.CREATION_IN_PROGRESS,
    JobStatus.IN_QUEUE,
    JobStatus.ON_VALIDATION,
    JobStatus.RUNNING,
]

METADATA_LIST = "metadataList"


class MetadataFields(StrEnum):
    CUSTOMER_NAME: str = "customer_name"
    PREP_CATEGORY: str = "prep_category"
    SAMPLE_NAME: str = "sample_name"
    SEQUENCED_AT: str = "sequenced_at"
    TICKET_NUMBER: str = "ticket_number"
    NAME = "metadataName"
    VALUE = "value"



import logging
from pathlib import Path

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.meta.workflow.downsample.sbatch import WORKFLOW_TEMPLATE
from cg.models.cg_config import CGConfig
from cg.store.models import Sample

LOG = logging.getLogger("logger")


class DownsampleWorkflow:
    def __init__(
        self,
        config: CGConfig,
        number_of_reads: int,
        downsampled_sample: Sample,
        dry_run: bool,
        output_fastq_dir: str,
        input_fastq_dir: str,
        original_sample: Sample,
        account: str = None,
    ):
        if dry_run:
            downsampled_sample = Sample()
            original_sample = Sample()

        self.config: CGConfig = config
        self.downsampled_sample: Sample = downsampled_sample
        self.number_of_reads: int = number_of_reads
        self.dry_run: bool = dry_run
        self.input_fastq_dir: str = input_fastq_dir
        self.original_sample: Sample = original_sample
        self.output_fastq_dir: str = output_fastq_dir
        self.slurm_api: SlurmAPI = SlurmAPI()
        self._account: str = account or self.config.downsample.account
        self._memory: str = "100"
        self._number_tasks: str = "2"
        self._time: str = "10:00:00"
        self._quality_of_service: str = "normal"
        self._email: str = "hiseq.clinical@scilifelab.se"

    @property
    def account(self) -> str:
        return self._account

    @property
    def email(self) -> str:
        return self._email

    @property
    def downsample_script_path(self) -> str:
        """The path to the downsample.sh script on Hasta."""
        return self.config.downsample.downsample_script

    @property
    def job_name(self) -> str:
        LOG.info(
            f"Jobname: downsample_{self.original_sample.internal_id}_to_{self.downsampled_sample.internal_id}"
        )
        return f"downsample_{self.original_sample.internal_id}_to_{self.downsampled_sample.internal_id}"

    @property
    def memory(self) -> str:
        LOG.info(f"Memory: {self._memory}")
        return self._memory

    @property
    def number_tasks(self) -> str:
        LOG.info(f"Number of tasks: {self._number_tasks}")
        return self._number_tasks

    @property
    def original_sample_reads(self) -> int:
        LOG.info(f"Original sample reads: {self.original_sample.reads}")
        return self.original_sample.reads

    @property
    def quality_of_service(self) -> str:
        LOG.info(f"Submitting job with qos: {self._quality_of_service}")
        return self._quality_of_service

    @property
    def sbatch_content(self) -> str:
        return WORKFLOW_TEMPLATE.format(
            account=self.account,
            bin_downsample=self.downsample_script_path,
            config_path=self.config,
            downsample_to=str(self.number_of_reads),
            bundle=self.downsampled_sample.internal_id if not self.dry_run else "dry_run",
            email=self.email,
            input_fastq_dir=self.input_fastq_dir,
            job_name=self.job_name,
            memory=self.memory,
            number_tasks=self.number_tasks,
            original_sample_reads=self.original_sample_reads,
            output_fastq_dir=self.output_fastq_dir,
            quality_of_service=self.quality_of_service,
            time=self.time,
        )

    @property
    def sbatch_name(self) -> str:
        return (
            f"{self.downsampled_sample.internal_id}_downsample.sh"
            if not self.dry_run
            else "dry_run_downsample.sh"
        )

    @property
    def time(self) -> str:
        LOG.info(f"Time for down sampling: {self._time}")
        return self._time

    def write_and_submit_sbatch_script(self) -> int:
        sbatch_path = Path(self.output_fastq_dir, self.sbatch_name)
        self.slurm_api.set_dry_run(dry_run=self.dry_run)

        return self.slurm_api.submit_sbatch(
            sbatch_content=self.sbatch_content,
            sbatch_path=sbatch_path,
        )



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.mutant.constants import MutantDeliveryType
from cg.services.orders.validation.order_types.mutant.models.sample import MutantSample


class MutantOrder(OrderWithSamples):
    delivery_type: MutantDeliveryType
    samples: list[MutantSample]

    @property
    def enumerated_samples(self) -> enumerate[MutantSample]:
        return enumerate(self.samples)



import logging
from pathlib import Path

from cg.io.txt import concat_txt, write_txt
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class NextflowConfigFileCreator:

    def __init__(
        self, account: str, platform: str, resources: str, store: Store, workflow_config_path: str
    ):
        self.account = account
        self.platform = platform
        self.resources = resources
        self.store = store
        self.workflow_config_path = workflow_config_path

    def create(self, case_id: str, file_path: Path) -> None:
        """Create the Nextflow config file for a case."""
        LOG.debug(f"Creating Nextflow config file for case {case_id}")
        content: str = self._get_content(case_id)
        write_txt(file_path=file_path, content=content)

    def _get_content(self, case_id: str) -> str:
        """Get the content of the Nextflow config file."""
        config_files_list: list[str] = [
            self.platform,
            self.workflow_config_path,
            self.resources,
        ]
        case_specific_params: list[str] = [
            self._get_cluster_options(case_id),
        ]
        return concat_txt(
            file_paths=config_files_list,
            str_content=case_specific_params,
        )

    def _get_cluster_options(self, case_id: str) -> str:
        case: Case = self.store.get_case_by_internal_id(case_id)
        return f'process.clusterOptions = "-A {self.account} --qos={case.slurm_priority}"\n'



"""API for uploading RAREDISEASE observations."""

import logging

from housekeeper.store.models import File, Version

from cg.constants.constants import CustomerId, SampleType
from cg.constants.observations import (
    LOQUSDB_ID,
    LOQUSDB_RARE_DISEASE_CUSTOMERS,
    LOQUSDB_RARE_DISEASE_SEQUENCING_METHODS,
    LoqusdbInstance,
    RarediseaseLoadParameters,
    RarediseaseObservationsAnalysisTag,
)
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.exc import CaseNotFoundError, LoqusdbDuplicateRecordError
from cg.meta.observations.observations_api import ObservationsAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.observations.input_files import RarediseaseObservationsInputFiles
from cg.store.models import Case
from cg.utils.dict import get_full_path_dictionary

LOG = logging.getLogger(__name__)


class RarediseaseObservationsAPI(ObservationsAPI):
    """API to manage RAREDISEASE observations."""

    def __init__(self, config: CGConfig):
        self.analysis_api = RarediseaseAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)
        self.loqusdb_api = None

    @property
    def loqusdb_customers(self) -> list[CustomerId]:
        """Customers that are eligible for RAREDISEASE Loqusdb uploads."""
        return LOQUSDB_RARE_DISEASE_CUSTOMERS

    @property
    def loqusdb_sequencing_methods(self) -> list[str]:
        """Sequencing methods that are eligible for Loqusdb uploads."""
        return LOQUSDB_RARE_DISEASE_SEQUENCING_METHODS

    @staticmethod
    def is_sample_type_eligible_for_observations_upload(case: Case) -> bool:
        """Return whether a RAREDISEASE case is free of tumor samples."""
        if case.tumour_samples:
            LOG.error(f"Sample type {SampleType.TUMOR} is not supported for Loqusdb uploads")
            return False
        return True

    def is_case_eligible_for_observations_upload(self, case: Case) -> bool:
        """Return whether a RAREDISEASE case is eligible for observations upload."""
        return all(
            [
                self.is_customer_eligible_for_observations_upload(case.customer.internal_id),
                self.is_sequencing_method_eligible_for_observations_upload(case.internal_id),
                self.is_sample_type_eligible_for_observations_upload(case),
                self.is_sample_source_type_not_ffpe(case.internal_id),
            ]
        )

    def set_loqusdb_instance(self, case_id: str) -> None:
        """Set the Loqusdb instance associated to the sequencing method."""
        sequencing_method: SeqLibraryPrepCategory = self.analysis_api.get_data_analysis_type(
            case_id
        )
        loqusdb_instances: dict[SeqLibraryPrepCategory, LoqusdbInstance] = {
            SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING: LoqusdbInstance.WGS,
            SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING: LoqusdbInstance.WES,
        }
        self.loqusdb_api = self.get_loqusdb_api(loqusdb_instances[sequencing_method])

    def load_observations(self, case: Case) -> None:
        """
        Load observation counts to Loqusdb for a case.
        Raises:
            LoqusdbDuplicateRecordError: If case has already been uploaded.
        """
        self.set_loqusdb_instance(case.internal_id)
        input_files: RarediseaseObservationsInputFiles = self.get_observations_input_files(case)
        if self.is_duplicate(
            case=case,
            loqusdb_api=self.loqusdb_api,
            profile_vcf_path=input_files.profile_vcf_path,
            profile_threshold=RarediseaseLoadParameters.PROFILE_THRESHOLD.value,
        ):
            LOG.error(
                f"Case {case.internal_id} has already been uploaded to {repr(self.loqusdb_api)}"
            )
            raise LoqusdbDuplicateRecordError

        load_output: dict = self.loqusdb_api.load(
            case_id=case.internal_id,
            snv_vcf_path=input_files.snv_vcf_path,
            sv_vcf_path=input_files.sv_vcf_path,
            profile_vcf_path=input_files.profile_vcf_path,
            family_ped_path=input_files.family_ped_path,
            gq_threshold=RarediseaseLoadParameters.GQ_THRESHOLD.value,
            hard_threshold=RarediseaseLoadParameters.HARD_THRESHOLD.value,
            soft_threshold=RarediseaseLoadParameters.SOFT_THRESHOLD.value,
        )
        loqusdb_id = str(self.loqusdb_api.get_case(case_id=case.internal_id)[LOQUSDB_ID])
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=loqusdb_id)
        LOG.info(f"Uploaded {load_output['variants']} variants to {repr(self.loqusdb_api)}")

    def get_observations_files_from_hk(
        self, hk_version: Version, case_id: str
    ) -> RarediseaseObservationsInputFiles:
        """Return observations files given a Housekeeper version for RAREDISEASE."""
        input_files: dict[str, File] = {
            "snv_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[RarediseaseObservationsAnalysisTag.SNV_VCF]
            ).first(),
            "sv_vcf_path": (
                self.housekeeper_api.files(
                    version=hk_version.id, tags=[RarediseaseObservationsAnalysisTag.SV_VCF]
                ).first()
                if self.analysis_api.get_data_analysis_type(case_id)
                == SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
                else None
            ),
            "profile_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[RarediseaseObservationsAnalysisTag.SNV_VCF]
            ).first(),
            "family_ped_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[RarediseaseObservationsAnalysisTag.FAMILY_PED]
            ).first(),
        }
        return RarediseaseObservationsInputFiles(**get_full_path_dictionary(input_files))

    def delete_case(self, case_id: str) -> None:
        """Delete RAREDISEASE case observations from Loqusdb."""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        self.set_loqusdb_instance(case_id)
        if not self.loqusdb_api.get_case(case_id):
            LOG.error(f"Case {case_id} could not be found in Loqusdb. Skipping case deletion.")
            raise CaseNotFoundError
        self.loqusdb_api.delete_case(case_id)
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=None)
        LOG.info(f"Removed observations for case {case_id} from {repr(self.loqusdb_api)}")



"""Unified interface to handle sample submissions.

This service will update information in Status and/or LIMS as required.

The normal entry for information is through the REST API which will pass a JSON
document with all information about samples in the submission. The input will
be validated and if passing all checks be accepted as new samples.
"""

from cg.models.orders.constants import OrderType
from cg.services.orders.storing.service import StoreOrderService
from cg.services.orders.storing.service_registry import StoringServiceRegistry
from cg.services.orders.submitter.ticket_handler import TicketHandler
from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.service import OrderValidationService
from cg.store.models import User


class OrderSubmitter:
    """Orders API for accepting new samples into the system."""

    def __init__(
        self,
        ticket_handler: TicketHandler,
        storing_registry: StoringServiceRegistry,
        validation_service: OrderValidationService,
    ):
        super().__init__()
        self.ticket_handler = ticket_handler
        self.storing_registry = storing_registry
        self.validation_service = validation_service

    def submit(self, order_type: OrderType, raw_order: dict, user: User) -> dict:
        """Submit a batch of samples.

        Main entry point for the class towards interfaces that implements it.
        """
        storing_service: StoreOrderService = self.storing_registry.get_storing_service(order_type)
        order: Order = self.validation_service.parse_and_validate(
            raw_order=raw_order, order_type=order_type, user_id=user.id
        )
        ticket_number: int = self.ticket_handler.create_ticket(
            order=order, user_name=user.name, user_mail=user.email, order_type=order_type
        )
        order._generated_ticket_id = ticket_number
        return storing_service.store_order(order)



"""CLI support to create config and/or start TAXPROFILER."""

import logging
from typing import cast

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, resolve_compression
from cg.cli.workflow.nf_analysis import (
    OPTION_RESUME,
    OPTION_REVISION,
    metrics_deliver,
    report_deliver,
    store,
    store_available,
    store_housekeeper,
)
from cg.constants.constants import MetaApis, Workflow
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.services.analysis_starter.analysis_starter import AnalysisStarter
from cg.services.analysis_starter.configurator.implementations.nextflow import NextflowConfigurator
from cg.services.analysis_starter.factories.configurator_factory import ConfiguratorFactory
from cg.services.analysis_starter.factories.starter_factory import AnalysisStarterFactory

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def taxprofiler(context: click.Context) -> None:
    """nf-core/taxprofiler analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = TaxprofilerAnalysisAPI(config=context.obj)


@taxprofiler.command()
@OPTION_REVISION
@ARGUMENT_CASE_ID
@click.pass_obj
def start(cg_config: CGConfig, case_id: str, revision: str | None) -> None:
    """Start a Taxprofiler case. \n
    Configures the case and writes the following files: \n
        - CASE_ID_params_file.yaml \n
        - CASE_ID_nextflow_config.json \n
        - CASE_ID_samplesheet.csv \n
    and submits the job to the Seqera Platform.
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.TAXPROFILER
    )
    analysis_starter.start(case_id=case_id, revision=revision)


@taxprofiler.command()
@click.pass_obj
def start_available(cg_config: CGConfig) -> None:
    """Starts all available Taxprofiler cases."""
    LOG.info("Starting Taxprofiler workflow for all available cases.")
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.TAXPROFILER
    )
    succeeded: bool = analysis_starter.start_available()
    if not succeeded:
        raise click.Abort


@taxprofiler.command()
@OPTION_REVISION
@OPTION_RESUME
@ARGUMENT_CASE_ID
@click.pass_obj
def run(cg_config: CGConfig, case_id: str, resume: bool, revision: str | None) -> None:
    """Run a preconfigured Taxprofiler case. \n
    Assumes that the following files are in the case run directory: \n
        - CASE_ID_params_file.yaml \n
        - CASE_ID_nextflow_config.json \n
        - CASE_ID_samplesheet.csv
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.TAXPROFILER
    )
    analysis_starter.run(case_id=case_id, resume=resume, revision=revision)


@taxprofiler.command()
@ARGUMENT_CASE_ID
@click.pass_obj
def config_case(cg_config: CGConfig, case_id: str) -> None:
    """Configure a Taxprofiler case so that it is ready to be run. \n
    Creates the following files in the case run directory:\n
        - CASE_ID_params_file.yaml \n
        - CASE_ID_nextflow_config.json \n
        - CASE_ID_samplesheet.csv
    """
    factory = ConfiguratorFactory(cg_config)
    configurator = cast(NextflowConfigurator, factory.get_configurator(Workflow.TAXPROFILER))
    configurator.configure(case_id=case_id)


taxprofiler.add_command(resolve_compression)
taxprofiler.add_command(metrics_deliver)
taxprofiler.add_command(report_deliver)
taxprofiler.add_command(store_housekeeper)
taxprofiler.add_command(store)
taxprofiler.add_command(store_available)
